#!/usr/bin/env python3
# Copyright 2016 The Font Bakery Authors.
# Copyright 2017 The Google Font Tools Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import argparse
import glob
import logging
from fontTools import ttLib


# set up some command line argument processing
parser = argparse.ArgumentParser(description="Print vendorID" " of TTF files")
parser.add_argument(
    "arg_filepaths",
    nargs="+",
    help="font file path(s) to check." " Wildcards like *.ttf are allowed.",
)


def main(args=None):
    # set up a basic logging config
    # to include timestamps
    # log_format = '%(asctime)s %(levelname)-8s %(message)s'
    global font
    log_format = "%(levelname)-8s %(message)s  "
    logger = logging.getLogger()
    handler = logging.StreamHandler()
    formatter = logging.Formatter(log_format)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    args = parser.parse_args(args)

    # ------------------------------------------------------
    logging.debug("Checking each file is a ttf")
    fonts_to_check = []
    for arg_filepath in sorted(args.arg_filepaths):
        # use glob.glob to accept *.ttf
        for fullpath in glob.glob(arg_filepath):
            file_path, file_name = os.path.split(fullpath)
            if file_name.endswith(".ttf"):
                fonts_to_check.append(fullpath)
            else:
                logging.warning(
                    "Skipping '{}' as it does not seem "
                    "to be valid TrueType font file.".format(file_name)
                )
    fonts_to_check.sort()

    if fonts_to_check == []:
        logging.error("None of the fonts are valid TrueType files!")

    # ------------------------------------------------------
    for font_file in fonts_to_check:
        font = ttLib.TTFont(font_file)
        logging.info("OK: {} opened with fontTools".format(font_file))

        # ----------------------------------------------------
        vid = font["OS/2"].achVendID
        print("[{}]: {}".format(vid, font_file))


__author__ = "The Google Fonts Tools Authors"
if __name__ == "__main__":
    main()
